//
//  engines.h
//  cubesql
//
//  Created by Marco Bambini on 10/02/15.
//
//

#ifndef __CUBESQL_ENGINES_H__
#define __CUBESQL_ENGINES_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif
	
#ifdef WIN32
#include <Windows.h>
#define CUBESQL_API __declspec(dllexport)
#define snprintf				_snprintf
#else
#define CUBESQL_API
#endif

#if defined(_MSC_VER) || defined(__BORLANDC__)
typedef __int64			cubesql_int64;
#else
typedef long long int	cubesql_int64;
#endif

#define CUBESQL_ENGINE_VERSION			1
#define kFALSE							0
#define kTRUE							1
#define PATHTYPE_ENGINE					28
#define PATHTYPE_LEN					4096
#define NULL_VALUE						-1
#define kExternalEngineError			7100
#define	kMemError						7003

#define DEBUG_WRITE(...)				printf(__VA_ARGS__);printf("\n");fflush(stdout)
	
// column types coming from the server
enum {
	COLTYPE_NONE				= 0,
	COLTYPE_INTEGER				= 1,
	COLTYPE_FLOAT				= 2,
	COLTYPE_TEXT				= 3,
	COLTYPE_BLOB				= 4,
	COLTYPE_BOOLEAN				= 5,
	COLTYPE_DATE				= 6,
	COLTYPE_TIME				= 7,
	COLTYPE_TIMESTAMP			= 8,
	COLTYPE_CURRENCY			= 9
};

typedef struct extern_resolver {
	// client
	void*	(*client_getdb) (void *client);
	void	(*client_seterror) (void *client, int errcode, char *message);
	void	(*client_set_affectedrows) (void *client, cubesql_int64 nrows);
	void	(*client_set_lastrowid) (void *client, cubesql_int64 rowid);
	int		(*client_senderror) (void *client);
	int		(*client_send_cursor) (void *client, void *cursor);
	
	// cursor
	void*	(*cursor_allocate) (int nrows, int ncols);
	void	(*cursor_free) (void *cursor);
	int		(*cursor_copy_char) (void *cursor, char *data, int len);
	int		(*cursor_copy_int) (void *curosr, int value);
	void	(*cursor_set_nrows) (void *cursor, int nrows);
	
	// cubesql
	void	(*cubesql_getpath) (char *path, int type);
	void	(*cubesql_pathcombine) (char *path, const char *path1, const char *path2);
	
} extern_resolver;

// function prototypes
// client
void *client_getdb (void *client);
void client_set_error (void *client, int errcode, char *message);
void client_set_affectedrows (void *client, cubesql_int64 nrows);
void client_set_lastrowid (void *client, cubesql_int64 rowid);
int  client_senderror (void *client);
int  client_sendcursor (void *client, void *cursor);

// cursor
void *cursor_allocate (int nrows, int ncols);
void cursor_free (void *cursor);
int  cursor_copy_char (void *cursor, char *data, int len);
int  cursor_copy_int (void *cursor, int value);
void cursor_set_nrows (void *curosr, int nrows);

// cubesql
void cubesql_getpath (char *path, int type);
void cubesql_pathcombine (char *path, const char *path1, const char *path2);
	
#ifdef __cplusplus
}
#endif

#endif
